import * as AWS from 'aws-sdk';

export class CheckJobs {
  private ec2 = new AWS.EC2();

  private readonly MINUTE_IN_SECONDS = 60;
  private readonly HOUR_IN_SECONDS = 60 * this.MINUTE_IN_SECONDS;
  private readonly DAY_IN_SECONDS = 24 * this.HOUR_IN_SECONDS;

  async getInstances() {
    const filters: AWS.EC2.Filter[] = [{
      Name: 'tag:LambdaFunction',
      Values: ['nshmp-haz-s3-lambda']
    }];

    const results = await this.ec2.describeInstances({Filters: filters}).promise();
    return this.processReservations(results.Reservations);
  }

  private processReservations(reservations: AWS.EC2.Reservation[]): Result {
    const responses: ResponseData[] = [];
    console.log('\n Reservations: ', JSON.stringify(reservations, null, 2));

    const users = this.getUsers(reservations);

    users.forEach(user => {
      const userReservations = reservations.filter(reservation => {
        return reservation.Instances.find(instance => {
          return this.getTag(instance, 'User').Value === user;
        });
      });

      const jobs: RunningJob[] = [];

      userReservations.forEach(reservation => {
        reservation.Instances.forEach(instance => {
          const job: RunningJob = {
            jobId: this.getTag(instance, 'JobId').Value,
            triggerFile: this.getTag(instance, 'TriggerFile').Value,
            state: instance.State.Name,
            instanceType: instance.InstanceType,
            launchTime: this.getLaunchTime(instance),
            s3OutputPath: this.getTag(instance, 'nshmp:s3-output-path').Value,
            configFile: this.getTag(instance, 'nshmp:config-file').Value,
            siteFile: this.getTag(instance, 'nshmp:site-file').Value
          };

          this.checkRunning(instance, job);
          jobs.push(job);
        });
      });

      responses.push({
        user: user,
        jobs
      });
    });

    return {
      status: 'success',
      results: responses
    }
  }

  private getUsers(reservations: AWS.EC2.Reservation[]): Set<string> {
    const users = new Set<string>();
    reservations.forEach(reservation => {
      reservation.Instances.forEach(instance => {
        users.add(this.getTag(instance, 'User').Value);
      });
    });

    return users;
  }

  private checkRunning(instance: AWS.EC2.Instance, job: RunningJob): void {
    if (instance.State.Name == 'running') {
      const now = new Date();
      const diffInSeconds = new Date(now.getTime() - instance.LaunchTime.getTime())
          .getTime() / 1000;
      let runTime = '';

      if (diffInSeconds >= this.DAY_IN_SECONDS) {
        runTime = `${(diffInSeconds / this.DAY_IN_SECONDS).toFixed(4)} days`;
      } else if (diffInSeconds >= this.HOUR_IN_SECONDS && diffInSeconds < this.DAY_IN_SECONDS) {
        runTime = `${(diffInSeconds / this.HOUR_IN_SECONDS).toFixed(4)} hours`;
      } else {
        runTime = `${(diffInSeconds / this.MINUTE_IN_SECONDS).toFixed(4)} minutes`;
      }

      job.runTime = runTime;
    }
  }

  private getLaunchTime(instance: AWS.EC2.Instance): string {
    const date = instance.LaunchTime.toLocaleDateString();
    const timeGMT = instance.LaunchTime.toLocaleTimeString();
    const timeMST = instance.LaunchTime
        .toLocaleTimeString('en-US', {timeZone: 'America/Denver'});
    return `${date} ${timeGMT} GMT (${timeMST} MST)`;
  }

  private getTag(instance: AWS.EC2.Instance, key: string): AWS.EC2.Tag {
    return instance.Tags.find(tag => tag.Key == key);
  }

}

interface Result {
  results: ResponseData[];
  status: string;
}

interface ResponseData {
  user: string;
  jobs: RunningJob[];
}

interface RunningJob {
  instanceType: string;
  jobId: string;
  launchTime: string;
  state: string;
  configFile: string;
  s3OutputPath: string;
  siteFile: string;
  triggerFile: string;
  runTime?: string;
}
