import * as AWS from 'aws-sdk';
import { PromiseResult } from 'aws-sdk/lib/request';
import * as dateFormat from 'dateformat';

import { LambdaConfig } from '../config/lambda-config';

/**
 * Lambda utilities
 */
export class NshmpLambdaUtil {

  static timeFormat = 'dddd, mmmm dS, yyyy, h:MM:ss TT Z';

  private constructor() {}

  /**
   * Send error message notification.
   *
   * @param config The config
   * @param errorMessage The error message to send
   */
  static onError(lambdaError: LambdaError):
      Promise<PromiseResult<AWS.SNS.PublishResponse, AWS.AWSError> | void> {
    const config = lambdaError.lambdaConfig;

    console.error(lambdaError);

    let message = `
      AWS Lambda nshmp-haz error.
      ${dateFormat(new Date(), NshmpLambdaUtil.timeFormat)}

      ----------
      Error:

      ${lambdaError.message}

      ----------
      Trigger config file (${config.triggerFileName}):
      \n${config.rawTriggerConfig}

    `;

    const subject = `AWS nshmp-haz (job id: ${config.jobId})`;

    return NshmpLambdaUtil.sendNotification(subject, message, config);
  }

  static sendNotification(subject: string, message: string, config: LambdaConfig):
      Promise<PromiseResult<AWS.SNS.PublishResponse, AWS.AWSError> | void> {
    const sns = new AWS.SNS();

    if (config.nshmp.snsTopic.TopicArn === null) {
      return Promise.reject();
    }

    const topicArn = config.nshmp.snsTopic.TopicArn;

    return sns.publish({
      TopicArn: topicArn,
      Message: message,
      Subject: subject
    }).promise()
    .catch((error: Error) => {
      const errorMessage = `
        Could not send notification.
        Error: ${error.message}.
        Topic Arn: ${topicArn},
        Subject: ${subject}.
        Message: ${message}.
      `;

      console.error(errorMessage);
    });
  }

}

export class LambdaError extends Error {

  lambdaConfig: LambdaConfig;

  constructor(message: string, lambdaConfig: LambdaConfig) {
    super(message);
    this.lambdaConfig = lambdaConfig;
  }

}
